#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <aversive.h>
#include <aversive/endian.h>
#include <aversive/wait.h>
#include <callout.h>
#include <uart.h>

#include "main.h"
#include "gps_venus.h"
#include "fat.h"
#include "fat_config.h"
#include "partition.h"
#include "sd_raw.h"
#include "sd_raw_config.h"
#include "sd_log.h"

#define debug_printf(args...) do { } while (0)
/* #define debug_printf(args...) printf(args) */

#define GPS_UART 0

/*
  https://code.google.com/p/hardware-ntp/source/browse/trunk/pcb-1/venus634.c?r=12
 */

/* some useful defines */
#define START1 0xA0
#define START2 0xA1
#define END1 0x0D
#define END2 0x0A

/* proccessing states */

#define S_RESET 0
#define S_POSTRESET 1
#define S_START1 2
#define S_START2 3
#define S_LENGTH1 4
#define S_LENGTH2 5
#define S_DISCARD 6
#define S_COPY 7
#define S_ENDCOPY 8
#define S_END 9
#define S_END1 10

#define MAX_LEN 0x200
struct gps_frame {
	uint16_t len;
	uint16_t cur_len;
	uint8_t data[MAX_LEN];
};

static struct gps_frame rxframe;

static struct gps_pos gps_pos;

static struct callout gps_timer;

/* INPUT FORMATS */

/* packet types */
/* 0x01 - System Restart */
#define M_RESTART 0x01
typedef struct {
	uint8_t start_mode; /* 01 = Hot; 02 = Warm; 03 = Cold */
	uint16_t utc_year; /* >= 1980 */
	uint8_t utc_month;
	uint8_t utc_day;
	uint8_t utc_hour;
	uint8_t utc_minute;
	uint8_t utc_second;
	int16_t latitude; /* -9000 to 9000, 1/100th degree, positive north */
	int16_t longitude; /* -18000 to 18000, 1/100th degree, positive east */
	int16_t altitude; /* -1000 to 18300, meters */
} m_restart_t;


/* 0x02 - Query Software Version */
#define M_QUERYVER 0x02
typedef struct {
	uint8_t software_type; /* 01 = System Code */
} m_queryver_t;


/* 0x05 - Serial Port Configuration */
#define M_SERCONF 0x05
typedef struct {
	uint8_t port; /* 00 = COM1 (the only port) */
	uint8_t baud; /* 00 = 4800; 01 = 9600; 02 = 19200; 03 = 38400;
			 04 = 57600; 05 = 115200; */
	uint8_t update; /* 00 = SRAM only; 01 = SRAM and FLASH */
} m_serconf_t;



/* 0x08 - nmea interval */
#define M_NMEAINTERVAL 0x08
typedef struct {
	uint8_t gga;
	uint8_t gsa;
	uint8_t gsv;
	uint8_t gll;
	uint8_t rmc;
	uint8_t vtg;
	uint8_t zda;
	uint8_t attributes; /*
			      0 - sram
			      1 - flash
			     */

} m_nmeainterval_t;


/* 0x09 - Query Software Version */
#define M_OUTPUT 0x09
typedef struct {
	uint8_t msg_type; /*
			    00 = no output
			    01 = nmea output
			    02 = binary output
			   */
} m_output_t;


/* 0x0E - Update Rage */
#define M_RATE 0x0E
typedef struct {
	uint8_t rate; /* Hz */
	uint8_t attributes; /*
			      0 - update ram
			      1 - update ram & flash
			    */
} m_rate_t;



/* 0x37 - configure waas */
#define M_WAAS 0x37
typedef struct {
	uint8_t enable; /* 01 = enable */
	uint8_t attributes;
} m_waas_t;

/* 0x3C - nav mode */
#define M_NAVMODE 0x3c
typedef struct {
	uint8_t navmode; /* 00 = car
			     01 = pedestrian
			   */
	uint8_t attributes;
} m_navmode_t;


#define DBG_SEND
#define DBG_RECV

void serial1_tx_cout(uint8_t c)
{
	debug_printf("%.2X ", c);
	uart_send(GPS_UART, c);
}

void venus634_send(uint8_t type, void *payload, uint16_t len)
{
	uint8_t crc = 0, n;

	debug_printf("SEND ");

	/* now send the message */
	/* header */
	serial1_tx_cout(START1);
	serial1_tx_cout(START2);
	serial1_tx_cout(((len+1) >> 8) & 0xff);
	serial1_tx_cout((len+1) & 0xff);
	/* type and payload */
	serial1_tx_cout(type);
	crc ^= type;
	for (n = 0; n < len; n++) {
		serial1_tx_cout(*(uint8_t *)payload);
		crc ^= *(uint8_t *)payload;
		payload++;
	}
	/* checksum and tail */
	serial1_tx_cout(crc);
	serial1_tx_cout(END1);
	serial1_tx_cout(END2);

	debug_printf("\n");
}

void venus634_restart(void)
{
	m_restart_t restart;

	memset(&restart,0,sizeof(m_restart_t));
	restart.start_mode = 0x03; /* COLD */

	venus634_send(M_RESTART,&restart,sizeof(m_restart_t));

	return;
}

void venus634_config_serial(void)
{
	m_serconf_t serconf;

	memset(&serconf,0,sizeof(m_serconf_t));
	serconf.port = 0;
	serconf.baud = 4;
	serconf.update = 1;

	venus634_send(M_SERCONF,&serconf,sizeof(m_serconf_t));

	return;
}


void venus634_msg_type(void)
{
	m_output_t output;

	memset(&output,0,sizeof(m_output_t));
	output.msg_type = 0x02; /* binary msg */

	venus634_send(M_OUTPUT,&output,sizeof(m_output_t));

	return;
}


void venus634_rate(void)
{
	m_rate_t rate;

	memset(&rate,0,sizeof(m_rate_t));
	rate.rate = 20;
	rate.attributes = 0; /* ram */

	venus634_send(M_RATE,&rate,sizeof(m_rate_t));

	return;
}

/* Wide Area Augmentation System: use ground stations to increase the precision
 * of gps position */
void venus634_waas(void)
{
	m_waas_t waas;

	memset(&waas,0,sizeof(m_waas_t));
	waas.enable = 1;
	waas.attributes = 0; /* ram */

	venus634_send(M_WAAS,&waas,sizeof(m_waas_t));

	return;
}

/* Tell we are a car instead of a pedestrian */
void venus634_navmode(void)
{
	m_navmode_t navmode;

	memset(&navmode,0,sizeof(m_navmode_t));
	navmode.navmode = 0; /* car */
	navmode.attributes = 0; /* ram */

	venus634_send(M_NAVMODE,&navmode,sizeof(m_navmode_t));

	return;
}

/* frequency of NMEA messages */
void venus634_nmea_interval(void)
{
	m_nmeainterval_t nmeainterval;

	memset(&nmeainterval,0,sizeof(m_nmeainterval_t));

	/* set frequency for nmea: 1 = once every one position fix, 2 = once
	 * every two position fix, ... */
	nmeainterval.gga = 1; /* GPGGA interval - GPS Fix Data*/
	nmeainterval.gsa = 1; /* GNSS DOPS and Active Satellites */
	nmeainterval.gsv = 1; /* GNSS Satellites in View */
	nmeainterval.gll = 1; /* Geographic Position - Latitude longitude */
	nmeainterval.rmc = 1; /* Recomended Minimum Specific GNSS Sentence */
	nmeainterval.vtg = 1; /* Course Over Ground and Ground Speed */
	nmeainterval.zda = 1; /* Time & Date*/

	nmeainterval.attributes = 1; /* ram flash */

	venus634_send(M_NMEAINTERVAL,&nmeainterval,sizeof(m_nmeainterval_t));

	return;
}

int8_t recv_cb(uint8_t byte)
{
	uint16_t i;

	/* bytes 0 and 1 are start bytes */
	if (rxframe.cur_len == 0) {
		if (byte != START1) {
			debug_printf("bad start1 %.2X\n", byte);
			goto reset_buf;
		}
	}
	else if (rxframe.cur_len == 1) {
		if (byte != START2) {
			debug_printf("bad start2 %.2X\n", byte);
			goto reset_buf;
		}
	}
	/* bytes 2 and 3 are the length of frame in network order */
	else if (rxframe.cur_len == 2) {
		rxframe.len = (uint16_t)byte << 8;
	}
	else if (rxframe.cur_len == 3) {
		rxframe.len |= (uint16_t)byte;
		if (rxframe.len > MAX_LEN) {
			debug_printf("bad len %d\n", rxframe.len);
			goto reset_buf;
		}
	}
	/* next bytes are data (the 4 below is the size of header) */
	else if ((rxframe.cur_len - 4) < rxframe.len) {
		rxframe.data[rxframe.cur_len - 4] = byte;
	}
	/* then it's the crc */
	else if ((rxframe.cur_len - 4) == rxframe.len) {
		uint8_t crc = 0;

		for (i = 0; i < rxframe.len; i++)
			crc ^= rxframe.data[i];
		if (byte != crc) {
			debug_printf("invalid crc\n");
			goto reset_buf;
		}

	}
	/* and the last 2 bytes are stop bytes */
	else if ((rxframe.cur_len - 4) == (rxframe.len + 1)) {
		if (byte != END1) {
			debug_printf("bad end1 %.2X\n", byte);
			goto reset_buf;
		}
	}
	else if ((rxframe.cur_len - 4) == (rxframe.len + 2)) {
		if (byte != END2) {
			debug_printf("bad end2 %.2X\n", byte);
			goto reset_buf;
		}
		debug_printf("valid frame received\n");
		rxframe.cur_len = 0;
		return 0;
	}
	else /* should not happen */
		goto reset_buf;

	rxframe.cur_len ++;
	return 1;

 reset_buf:
	rxframe.cur_len = 0;
	return 1;
}

int recv_msg(void)
{
	int ret;
	int16_t c;

	while (1) {
		/* XXX use select for timeout */
		c = uart_recv(GPS_UART);
		ret = recv_cb(c);
		if (ret == 0)
			return 0;
	}
}

int wait_ack(int msg_type)
{
	int ret;

	while (1) {
		ret = recv_msg();
		if (ret < 0)
			return -1;

		/* retry if it's not the expected message */
		if (rxframe.data[0] != 0x83 && rxframe.data[0] != 0x84)
			continue;
		if (rxframe.data[1] != msg_type)
			continue;

		if (rxframe.data[0] == 0x83)
			printf("ACK\n");
		else if (rxframe.data[0] == 0x84)
			printf("NACK\n");
		else
			printf("ZARB\n");
		break;
	}

	return 0;
}

static int decode_gps_pos(uint8_t *buf, uint16_t len)
{
	struct gps_pos *pos = (struct gps_pos *)buf;
	uint8_t irq_flags;

	if (len != sizeof(*pos))
		return -1;

	if (pos->msg_id != 0xA8) /* XXX not tested */
		return -1;

	gps_pos.gps_week = ntohs(pos->gps_week);
	gps_pos.tow = ntohl(pos->tow);

	gps_pos.latitude = ntohl(pos->latitude);
	gps_pos.longitude = ntohl(pos->longitude);
	gps_pos.altitude = ntohl(pos->altitude);

	gps_pos.sea_altitude = ntohl(pos->sea_altitude);

	gps_pos.gdop = ntohs(pos->gdop);
	gps_pos.pdop = ntohs(pos->pdop);
	gps_pos.hdop = ntohs(pos->hdop);
	gps_pos.vdop = ntohs(pos->vdop);
	gps_pos.tdop = ntohs(pos->tdop);

	gps_pos.ecef_vx = ntohl(pos->ecef_vx);
	gps_pos.ecef_vy = ntohl(pos->ecef_vy);
	gps_pos.ecef_vz = ntohl(pos->ecef_vz);

	/* update global structure */
	IRQ_LOCK(irq_flags);
	memcpy(&gps_pos, pos, sizeof(gps_pos));
	IRQ_UNLOCK(irq_flags);

	return 0;
}

/* display current GPS position stored in the global variable */
static void display_gps(void)
{
	printf("id %.2X mode %.2X svnum %.2X gpsw %.4X tow %.10"PRIu32"\t",
		gps_pos.msg_id,
		gps_pos.mode,
		gps_pos.sv_num,
		gps_pos.gps_week,
		gps_pos.tow);

	printf("lat %.8"PRIx32" long %.8"PRIx32" alt %.8"PRIx32"\n",
		gps_pos.latitude,
		gps_pos.longitude,
		gps_pos.altitude);

	printf("gdop %3.3f pdop %3.3f hdop %3.3f vdop %3.3f tdop %3.3f\n",
		(double)gps_pos.gdop/100.,
		(double)gps_pos.pdop/100.,
		(double)gps_pos.hdop/100.,
		(double)gps_pos.vdop/100.,
		(double)gps_pos.tdop/100.);

	printf("lat %3.5f long %3.5f alt %3.5f sea_alt %3.5f\n",
		(double)gps_pos.latitude/10000000.,
		(double)gps_pos.longitude/10000000.,
		(double)gps_pos.altitude/100.,
		(double)gps_pos.sea_altitude/100.);

	printf("vx %3.3f vy %3.3f vz %3.3f\n",
		(double)gps_pos.ecef_vx/100.,
		(double)gps_pos.ecef_vy/100.,
		(double)gps_pos.ecef_vz/100.);
}

static void gps_venus_cb(struct callout_mgr *cm, struct callout *tim, void *arg)
{
	int16_t c;
	int ret;

	(void)cm;
	(void)tim;
	(void)arg;

	while (1) {
		c = uart_recv_nowait(GPS_UART);
		if (c < 0) /* no more char */
			goto resched;

		ret = recv_cb(c);
		if (ret == 0) {
			decode_gps_pos(rxframe.data, rxframe.len);
			if (0)
				display_gps();
		}
	}

 resched:
	callout_schedule(cm, tim, 2);
}

static void venus634_configure(void)
{
	/* ask the GPS to reset */
	printf("init...");
	venus634_restart();
	wait_ack(M_RESTART);

	/* it seems we must wait that the GPS is restarted, else it doesn't work
	 * properly */
	wait_ms(500);

	printf("binmode...");
	venus634_msg_type();
	wait_ack(M_OUTPUT);

	printf("waas...");
	venus634_waas();
	wait_ack(M_WAAS);

	printf("rate...");
	venus634_rate();
	wait_ack(M_RATE);

	printf("GPS configuration done !\n");
}

/*
  https://www.sparkfun.com/datasheets/GPS/Modules/AN0003_v1.4.14_FlashOnly.pdf
*/

int gps_venus_init(void)
{
	venus634_configure();

	callout_init(&gps_timer, gps_venus_cb, NULL, GPS_PRIO);
	callout_schedule(&imuboard.intr_cm, &gps_timer, 2); /* every 2ms */

	return 0;
}

void gps_get_pos(struct gps_pos *pos)
{
	memcpy(pos, &gps_pos, sizeof(*pos));
}

int gps_loop(void)
{
	uint32_t ms;
	uint8_t flags, prio;
	int16_t len;
	char buf[128];
	struct gps_pos pos;

	while (1) {

		IRQ_LOCK(flags);
		ms = global_ms;
		IRQ_UNLOCK(flags);

		/* get position (prevent modification of gps pos during copy) */
		prio = callout_mgr_set_prio(&imuboard.intr_cm, GPS_PRIO);
		gps_get_pos(&pos);
		callout_mgr_restore_prio(&imuboard.intr_cm, prio);

		/* XXX copy */
		len = snprintf(buf, sizeof(buf),
			"%"PRIu32" "
			"svnum %.2X lat %3.5f long %3.5f "
			"alt %3.5f sea_alt %3.5f\n",
			ms, gps_pos.sv_num,
			(double)gps_pos.latitude/10000000.,
			(double)gps_pos.longitude/10000000.,
			(double)gps_pos.altitude/100.,
			(double)gps_pos.sea_altitude/100.);


		if (sd_log_enabled()) {

			if (sd_log_write(buf, len) != len) {
				printf_P(PSTR("error writing to file\n"));
				return -1;
			}
		}
		else
			printf("%s", buf);

	}

	return 0;
}
